import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.memory import (
    MAX,
    MIN,
    PROP_CONSISTENCY,
    PROP_ENTAILMENT,
    PROP_INCONSISTENCY,
    new_triggers,
)


def get_triggers(n: int, data: NDArray) -> NDArray:
    """
    Returns the triggers for this propagator.
    :param n: the number of variables
    :return: an array of triggers
    """
    triggers = new_triggers(n, False)
    for i in range(n - 1):
        triggers[i, MAX] = True
    triggers[-1, MIN] = True
    return triggers


@njit("int64(int32[::1,:], int32[:])", cache=True)
def compute_domains(domains: NDArray, data: NDArray) -> int:
    """
    Implements Min_i x_i >= x_{n-1}.
    :param domains: the domains of the variables
    :param data: unused here
    """
    x = domains[:-1]
    y = domains[-1]
    if y[MAX] <= np.min(x[:, MIN]):
        return PROP_ENTAILMENT
    y[MAX] = min(y[MAX], np.min(x[:, MAX]))
    if y[MIN] > y[MAX]:
        return PROP_INCONSISTENCY
    for i in range(len(x)):
        x[i, MIN] = max(x[i, MIN], y[MIN])
        if x[i, MAX] < x[i, MIN]:
            return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.memory import MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY, new_triggers


def get_triggers(n: int, data: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :return: an array of triggers
    """
    return new_triggers(n, True)


@njit(cache=True)
def compute_domain_sum(n: int, domains: NDArray, data: NDArray) -> NDArray:
    domain_sum = np.empty(2, dtype=np.int32)
    domain_sum[:] = data[-1]
    for i in range(n):
        c = data[i]
        if c > 0:
            domain_sum[MIN] -= domains[i, MAX] * c
            domain_sum[MAX] -= domains[i, MIN] * c
        elif c < 0:
            domain_sum[MIN] -= domains[i, MIN] * c
            domain_sum[MAX] -= domains[i, MAX] * c
    return domain_sum


@njit("int64(int32[::1,:], int32[:])", cache=True)
def compute_domains(domains: NDArray, a: NDArray) -> int:
    """
    Implements Sigma_i a_i * x_i = a_{n-1}.
    :param domains: the domains of the variables
    :param a: the parameters of the propagator
    """
    n = len(domains)
    domain_sum = compute_domain_sum(n, domains, a)
    new_domains = np.empty_like(domains)
    new_domains[:] = domains[:]
    for i in range(n):
        if a[i] > 0:
            new_min = domains[i, MAX] - (domain_sum[MIN] // -a[i])
            new_max = domains[i, MIN] + (domain_sum[MAX] // a[i])
        elif a[i] < 0:
            new_min = domains[i, MAX] - (-domain_sum[MAX] // a[i])
            new_max = domains[i, MIN] + (-domain_sum[MIN] // -a[i])
        new_domains[i, MIN] = max(domains[i, MIN], new_min)
        new_domains[i, MAX] = min(domains[i, MAX], new_max)
        if new_domains[i, MIN] > new_domains[i, MAX]:
            return PROP_INCONSISTENCY
    domains[:] = new_domains[:]
    return PROP_CONSISTENCY



import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.memory import MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY, new_triggers


def get_triggers(n: int, data: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :return: an array of triggers
    """
    return new_triggers(n, True)


@njit("(uint16[:], uint16, uint16, uint16)", cache=True)
def path_set(t: NDArray, start: int, end: int, to: int) -> None:
    while (p := start) != end:
        start = t[p]
        t[p] = to


@njit("uint16(uint16[:], uint16)", cache=True)
def path_min(t: NDArray, i: int) -> int:
    while t[i] < i:
        i = t[i]
    return i


@njit("uint16(uint16[:], uint16)", cache=True)
def path_max(t: NDArray, i: int) -> int:
    while t[i] > i:
        i = t[i]
    return i


@njit("uint16(int64, int32[::1, :], uint16[:, :], int64[:], int64[:], int32[:])", cache=True)
def compute_nb(
    n: int,
    domains: NDArray,
    ranks: NDArray,
    min_sorted_vars: NDArray,
    max_sorted_vars: NDArray,
    bounds: NDArray,
) -> int:
    min_value = domains[min_sorted_vars[0], MIN]
    max_value = domains[max_sorted_vars[0], MAX] + 1
    last = min_value - 2
    bounds[0] = last
    i = j = nb = 0
    while True:
        if i < n and min_value <= max_value:
            if min_value != last:
                nb += 1
                bounds[nb] = last = min_value
            ranks[min_sorted_vars[i], MIN] = nb
            i += 1
            if i < n:
                min_value = domains[min_sorted_vars[i], MIN]
        else:
            if max_value != last:
                nb += 1
                bounds[nb] = last = max_value
            ranks[max_sorted_vars[j], MAX] = nb
            j += 1
            if j == n:
                break
            max_value = domains[max_sorted_vars[j], MAX] + 1
    bounds[nb + 1] = bounds[nb] + 2
    return nb


@njit(
    "boolean(int64, int64, uint16[:], int32[:], uint16[:], int32[:], int32[::1, :], uint16[:, :], int64[:])",
    cache=True,
)
def filter_lower(
    n: int,
    nb: int,
    t: NDArray,
    d: NDArray,
    h: NDArray,
    bounds: NDArray,
    domains: NDArray,
    ranks: NDArray,
    max_sorted_vars: NDArray,
) -> bool:
    for i in range(1, nb + 2):
        t[i] = h[i] = i - 1
        d[i] = bounds[i] - bounds[i - 1]
    for i in range(n):
        max_sorted_vars_i = max_sorted_vars[i]
        x = ranks[max_sorted_vars_i, MIN]
        y = ranks[max_sorted_vars_i, MAX]
        z = path_max(t, x + 1)
        j = t[z]
        d[z] -= 1
        if d[z] == 0:
            t[z] = z + 1
            z = path_max(t, t[z])
            t[z] = j
        if d[z] + bounds[y] < bounds[z]:  # moved above the path compression which is not the case in the paper
            return False
        path_set(t, x + 1, z, z)  # path compression
        if h[x] > x:
            w = path_max(h, h[x])
            domains[max_sorted_vars_i, MIN] = bounds[w]
            path_set(h, x, w, w)  # path compression
        if d[z] + bounds[y] == bounds[z]:
            path_set(h, h[y], j - 1, y)  # mark hall interval
            h[y] = j - 1  # hall interval[bounds[j], bounds[y]]
    return True


@njit(
    "boolean(int64, int64, uint16[:], int32[:], uint16[:], int32[:], int32[::1, :], uint16[:, :], int64[:])",
    cache=True,
)
def filter_upper(
    n: int,
    nb: int,
    t: NDArray,
    d: NDArray,
    h: NDArray,
    bounds: NDArray,
    domains: NDArray,
    ranks: NDArray,
    min_sorted_vars: NDArray,
) -> bool:
    for i in range(nb + 1):
        t[i] = h[i] = i + 1
        d[i] = bounds[i + 1] - bounds[i]
    for i in range(n - 1, -1, -1):
        min_sorted_vars_i = min_sorted_vars[i]
        x = ranks[min_sorted_vars_i, MAX]
        y = ranks[min_sorted_vars_i, MIN]
        z = path_min(t, x - 1)
        j = t[z]
        d[z] -= 1
        if d[z] == 0:
            t[z] = z - 1
            z = path_min(t, t[z])
            t[z] = j
        if d[z] + bounds[z] < bounds[y]:  # moved above the path compression which is not the case in the paper
            return False
        path_set(t, x - 1, z, z)  # path compression
        if h[x] < x:
            w = path_min(h, h[x])
            domains[min_sorted_vars_i, MAX] = bounds[w] - 1
            path_set(h, x, w, w)  # path compression
        if d[z] + bounds[z] == bounds[y]:
            path_set(h, h[y], j + 1, y)  # mark hall interval
            h[y] = j + 1  # hall interval[bounds[j], bounds[y]]
    return True


@njit("int64(int32[::1,:], int32[:])", cache=True)
def compute_domains(domains: NDArray, data: NDArray) -> int:
    """
    Adapted from "A fast and simple algorithm for bounds consistency of the alldifferent constraint".
    :param domains: the domains of the variables
    :param data: unused here
    """
    n = len(domains)
    ranks = np.zeros((n, 2), dtype=np.uint16)
    bounds_nb = 2 * n + 2
    bounds = np.zeros(bounds_nb, dtype=np.int32)
    min_sorted_vars = np.argsort(domains[:, MIN])
    max_sorted_vars = np.argsort(domains[:, MAX])
    nb = compute_nb(n, domains, ranks, min_sorted_vars, max_sorted_vars, bounds)
    t = np.zeros(bounds_nb, dtype=np.uint16)  # critical capacity pointers
    d = np.zeros(bounds_nb, dtype=np.int32)  # differences between critical capacities
    h = np.zeros(bounds_nb, dtype=np.uint16)  # Hall interval pointers
    return (
        PROP_CONSISTENCY
        if filter_lower(n, nb, t, d, h, bounds, domains, ranks, max_sorted_vars)
        and filter_upper(n, nb, t, d, h, bounds, domains, ranks, min_sorted_vars)
        else PROP_INCONSISTENCY
    )



from typing import List

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AFFINE_EQ, ALG_ALLDIFFERENT

A = 0
B = 1
C = 2
D = 3
E = 4
F = 5
G = 6
H = 7
I = 8
J = 9
K = 10
L = 11
M = 12
N = 13
O = 14
P = 15
Q = 16
R = 17
S = 18
T = 19
U = 20
V = 21
W = 22
X = 23
Y = 24
Z = 25


class AlphaProblem(Problem):
    """
    This problem comes from the newsgroup rec.puzzle.
    The numbers from 1 to 26 are assigned to the letters of the alphabet.
    The numbers beside each word are the total of the values assigned to the letters in the word
    (e.g for LYRE: L,Y,R,E might be to equal 5,9,20 and 13 or any other combination that add up to 47).

    Find the value of each letter under the equations:
    BALLET  45     GLEE  66     POLKA      59     SONG     61
    CELLO   43     JAZZ  58     QUARTET    50     SOPRANO  82
    CONCERT 74     LYRE  47     SAXOPHONE 134     THEME    72
    FLUTE   30     OBOE  53     SCALE      51     VIOLIN  100
    FUGUE   50     OPERA 65     SOLO       37     WALTZ    34
    """

    def __init__(self) -> None:
        super().__init__(shr_domains=[(1, 26)] * 26, dom_indices=list(range(26)), dom_offsets=[0] * 26)
        self.set_propagators(
            [
                ([A, B, E, T, L], ALG_AFFINE_EQ, [1, 1, 1, 1, 2, 45]),
                ([C, E, O, L], ALG_AFFINE_EQ, [1, 1, 1, 2, 43]),
                ([E, O, N, R, T, C], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 2, 74]),
                ([E, F, L, U, T], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 30]),
                ([E, F, G, U], ALG_AFFINE_EQ, [1, 1, 1, 2, 50]),
                ([G, L, E], ALG_AFFINE_EQ, [1, 1, 2, 66]),
                ([A, J, Z], ALG_AFFINE_EQ, [1, 1, 2, 58]),
                ([E, L, R, Y], ALG_AFFINE_EQ, [1, 1, 1, 1, 47]),
                ([E, B, O], ALG_AFFINE_EQ, [1, 1, 2, 53]),
                ([A, E, P, O, R], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 65]),
                ([A, K, L, O, P], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 59]),
                ([A, E, Q, R, U, T], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 2, 50]),
                ([A, E, H, N, P, S, X, O], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 1, 1, 2, 134]),
                ([A, C, E, L, S], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 51]),
                ([L, S, O], ALG_AFFINE_EQ, [1, 1, 2, 37]),
                ([G, N, O, S], ALG_AFFINE_EQ, [1, 1, 1, 1, 61]),
                ([A, N, P, R, S, O], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 2, 82]),
                ([H, M, T, E], ALG_AFFINE_EQ, [1, 1, 1, 2, 72]),
                ([L, N, O, V, I], ALG_AFFINE_EQ, [1, 1, 1, 1, 2, 100]),
                ([A, L, T, W, Z], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 34]),
                ([A, B, C, D, E, F, G, H, I, J, K, L, M, N, O, P, Q, R, S, T, U, V, W, X, Y, Z], ALG_ALLDIFFERENT, []),
            ]
        )

    def pretty_print_solution(self, solution: List[int]) -> None:
        print(
            {
                "A": solution[A],
                "B": solution[B],
                "C": solution[C],
                "D": solution[D],
                "E": solution[E],
                "F": solution[F],
                "G": solution[G],
                "H": solution[H],
                "I": solution[I],
                "J": solution[J],
                "K": solution[K],
                "L": solution[L],
                "M": solution[M],
                "N": solution[N],
                "O": solution[O],
                "P": solution[P],
                "Q": solution[Q],
                "R": solution[R],
                "S": solution[S],
                "T": solution[T],
                "U": solution[U],
                "V": solution[V],
                "W": solution[W],
                "X": solution[X],
                "Y": solution[Y],
                "Z": solution[Z],
            }
        )



from typing import Tuple, Union

import numpy as np
from numba import njit  # type: ignore
from numba.typed import List
from numpy.typing import NDArray

from nucs.memory import (
    END,
    MAX,
    MIN,
    PROP_ENTAILMENT,
    PROP_INCONSISTENCY,
    START,
    new_algorithms,
    new_bounds,
    new_data,
    new_domain_offsets_by_values,
    new_domains_by_values,
    new_entailed_propagators,
    new_indices,
    new_indices_by_values,
    new_offsets,
    new_triggered_propagators,
    new_triggers,
)
from nucs.propagators.propagators import (
    compute_domains,
    get_triggers,
    init_triggered_propagators,
    update_triggered_propagators,
)
from nucs.statistics import (
    STATS_PROBLEM_FILTER_NB,
    STATS_PROPAGATOR_ENTAILMENT_NB,
    STATS_PROPAGATOR_FILTER_NB,
    STATS_PROPAGATOR_FILTER_NO_CHANGE_NB,
    STATS_PROPAGATOR_INCONSISTENCY_NB,
    statistics_init,
)


class Problem:
    """
    A problem is conceptually defined by a list of domains, a list of variables and a list of propagators.
    A domain can be computed by applying an offset to a shared domain.
    """

    def __init__(self, shr_domains: List[Union[int, Tuple[int, int]]], dom_indices: List[int], dom_offsets: List[int]):
        """
        Inits the problem.
        :param shr_domains: the shared domains expressed as a list
        :param dom_indices: the domain indices expressed as a list
        :param dom_offsets: the domain offsets expressed as a list
        """
        self.variable_nb = len(dom_indices)
        self.shr_domains_backup = shr_domains  # a copy of the shared domains (as a list) used when resetting a problem
        self.shr_domains = new_domains_by_values(shr_domains)
        self.dom_indices = new_indices_by_values(dom_indices)
        self.dom_offsets = new_domain_offsets_by_values(dom_offsets)
        self.statistics = statistics_init()

    def reset_shr_domains(self) -> None:
        """
        Resets the shared domains to their initial values.
        """
        self.shr_domains = new_domains_by_values(self.shr_domains_backup)

    def reset_entailed_propagators(self) -> None:
        """
        Marks all propagators as not entailed anymore.
        """
        self.entailed_propagators[:] = False

    def set_propagators(self, propagators: List[Tuple[List[int], int, List[int]]]) -> None:
        """
        Completes the initialization of the problem by setting the propagators.
        The order of the propagators has an impact on the global speed of the resolution:
        it is recommended to put the more costly propagators at the end of the list.
        :param propagators: the list of propagators as tuples of the form (list of variables, algorithm, data).
        """
        self.propagator_nb = len(propagators)
        props_var_total_size = props_data_total_size = 0
        self.triggered_propagators = new_triggered_propagators(self.propagator_nb)
        self.entailed_propagators = new_entailed_propagators(self.propagator_nb)
        self.algorithms = new_algorithms(self.propagator_nb)
        # We will store propagator specific data in a global arrays, we need to compute variables and data bounds.
        self.var_bounds = new_bounds(
            self.propagator_nb
        )  # there is a bit of redundancy here for faster access to the bounds
        self.data_bounds = new_bounds(
            self.propagator_nb
        )  # there is a bit of redundancy here for faster access to the bounds
        self.var_bounds[0, START] = self.data_bounds[0, START] = 0
        for pidx, propagator in enumerate(propagators):
            prop_var_size = len(propagator[0])
            prop_data_size = len(propagator[2])
            self.algorithms[pidx] = propagator[1]
            if pidx > 0:
                self.var_bounds[pidx, START] = self.var_bounds[pidx - 1, END]
                self.data_bounds[pidx, START] = self.data_bounds[pidx - 1, END]
            self.var_bounds[pidx, END] = self.var_bounds[pidx, START] + prop_var_size
            self.data_bounds[pidx, END] = self.data_bounds[pidx, START] + prop_data_size
            props_var_total_size += prop_var_size
            props_data_total_size += prop_data_size
        # Bounds have been computed and can now be used.
        # The global arrays are the following:
        self.props_dom_indices = new_indices(props_var_total_size)
        self.props_dom_offsets = new_offsets(props_var_total_size)
        self.props_triggers = new_triggers(props_var_total_size, False)
        self.props_data = new_data(props_data_total_size)
        # Let's init the global arrays.
        for pidx, propagator in enumerate(propagators):
            prop_vars = propagator[0]
            self.props_dom_indices[self.var_bounds[pidx, START] : self.var_bounds[pidx, END]] = self.dom_indices[
                prop_vars
            ]  # this is cached for faster access
            self.props_dom_offsets[self.var_bounds[pidx, START] : self.var_bounds[pidx, END]] = self.dom_offsets[
                prop_vars
            ]  # this is cached for faster access
            self.props_data[self.data_bounds[pidx, START] : self.data_bounds[pidx, END]] = propagator[2]
            self.props_triggers[self.var_bounds[pidx, START] : self.var_bounds[pidx, END]] = get_triggers(
                propagator[1], len(propagator[0]), propagator[2]
            )

    def get_values(self) -> List[int]:
        """
        Gets the values for the variables (when instantiated).
        :return: a list of integers
        """
        mins = self.shr_domains[self.dom_indices, MIN] + self.dom_offsets
        return mins.tolist()

    def set_min_value(self, var_idx: int, min_value: int) -> None:
        """
        Sets the minimal value of a variable.
        :param var_idx: the index of the variable
        :param min_value: the minimal value
        """
        self.shr_domains[(self.dom_indices[var_idx]), MIN] = min_value - self.dom_offsets[var_idx]

    def set_max_value(self, var_idx: int, max_value: int) -> None:
        """
        Sets the maximal value of a variable.
        :param var_idx: the index of the variable
        :param min_value: the maximal value
        """
        self.shr_domains[(self.dom_indices[var_idx]), MAX] = max_value - self.dom_offsets[var_idx]

    def __str__(self) -> str:
        # TODO: fix this
        return f"domains={self.shr_domains}"

    def filter(self, shr_domain_changes: NDArray) -> bool:
        """
        Filters the problem's domains by applying the propagators until a fix point is reached.
        :param shr_domain_changes: an array of shared domain changes
        :return: False if the problem is not consistent
        """
        return filter(
            self.statistics,
            self.propagator_nb,
            self.triggered_propagators,
            self.entailed_propagators,
            self.algorithms,
            self.var_bounds,
            self.data_bounds,
            self.props_dom_indices,
            self.props_dom_offsets,
            self.props_triggers,
            self.props_data,
            self.shr_domains,
            shr_domain_changes,
        )

    def pretty_print_solution(self, solution: List[int]) -> None:
        """
        Pretty prints a solution to the problem.
        :param solution: a list of integers
        """
        print(solution)


@njit(cache=True)
def pop_propagator(propagator_queue: NDArray) -> int:
    for prop_idx in range(len(propagator_queue)):
        if propagator_queue[prop_idx]:
            propagator_queue[prop_idx] = False
            return prop_idx
    return -1


@njit(cache=True)
def filter(
    statistics: NDArray,
    propagator_nb: int,
    triggered_propagators: NDArray,
    entailed_propagators: NDArray,
    algorithms: NDArray,
    var_bounds: NDArray,
    data_bounds: NDArray,
    props_indices: NDArray,
    props_offsets: NDArray,
    props_triggers: NDArray,
    props_data: NDArray,
    shr_domains: NDArray,
    shr_domain_changes: NDArray,
) -> bool:
    """
    Filters the problem's domains by applying the propagators until a fix point is reached.
    :param shr_domain_changes: an array of shared domain changes
    :return: False if the problem is not consistent
    """
    shr_domains_cur = np.empty_like(shr_domains)
    statistics[STATS_PROBLEM_FILTER_NB] += 1
    init_triggered_propagators(
        triggered_propagators,
        entailed_propagators,
        shr_domain_changes,
        propagator_nb,
        var_bounds,
        props_indices,
        props_triggers,
    )
    while True:
        prop_idx = pop_propagator(triggered_propagators)
        if prop_idx == -1:
            return True
        statistics[STATS_PROPAGATOR_FILTER_NB] += 1
        prop_var_start = var_bounds[prop_idx, START]
        prop_var_end = var_bounds[prop_idx, END]
        prop_indices = props_indices[prop_var_start:prop_var_end]
        prop_offsets = props_offsets[prop_var_start:prop_var_end].reshape((-1, 1))
        prop_domains = np.empty((2, len(prop_offsets)), dtype=np.int32).T  # trick for order=F
        np.add(shr_domains[prop_indices], prop_offsets, prop_domains)
        status = compute_domains(
            algorithms[prop_idx],
            prop_domains,
            props_data[data_bounds[prop_idx, START] : data_bounds[prop_idx, END]],
        )
        if status == PROP_INCONSISTENCY:
            statistics[STATS_PROPAGATOR_INCONSISTENCY_NB] += 1
            return False
        if status == PROP_ENTAILMENT:
            entailed_propagators[prop_idx] = True
            statistics[STATS_PROPAGATOR_ENTAILMENT_NB] += 1
        shr_domains_cur[:, :] = shr_domains
        shr_domains[prop_indices] = prop_domains - prop_offsets
        shr_domain_changes[:, :] = shr_domains_cur != shr_domains
        if np.any(shr_domain_changes):  # type: ignore
            update_triggered_propagators(
                triggered_propagators,
                entailed_propagators,
                shr_domain_changes,
                propagator_nb,
                var_bounds,
                props_indices,
                props_triggers,
                prop_idx,
            )
        else:
            statistics[STATS_PROPAGATOR_FILTER_NO_CHANGE_NB] += 1


@njit(cache=True)
def is_solved(shr_domains: NDArray) -> bool:
    """
    Returns true iff the problem is solved.
    :return: a boolean
    """
    return bool(np.all(np.equal(shr_domains[:, MIN], shr_domains[:, MAX])))



import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.memory import (
    MAX,
    MIN,
    PROP_CONSISTENCY,
    PROP_ENTAILMENT,
    PROP_INCONSISTENCY,
    new_triggers,
)


def get_triggers(n: int, data: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :return: an array of triggers
    """
    return new_triggers(n, True)


@njit("int64(int32[::1,:], int32[:])", cache=True)
def compute_domains(x: NDArray, data: NDArray) -> int:
    """
    Implements Sigma_i (x_i == a_0) = a_1.
    :param domains: the domains of the variables
    :param data: the parameters of the propagator
    """
    value = data[0]
    ok_count_max = len(x) - np.count_nonzero((x[:, MIN] > value) | (x[:, MAX] < value))
    ok_count_min = np.count_nonzero((x[:, MIN] == value) & (x[:, MAX] == value))
    counter = data[1]
    if ok_count_min > counter or ok_count_max < counter:
        return PROP_INCONSISTENCY
    if ok_count_min == counter and ok_count_max == counter:
        return PROP_ENTAILMENT
    if ok_count_min == counter:  # we cannot have more domains equal to c
        x[(x[:, MIN] == value) & (x[:, MAX] > value), MIN] = value + 1
        x[(x[:, MIN] < value) & (x[:, MAX] == value), MAX] = value - 1
    if ok_count_max == counter:  # we cannot have more domains different from c
        x[(x[:, MIN] <= value) & (value <= x[:, MAX]), :] = value
    return PROP_CONSISTENCY



import sys
from typing import Callable, List

from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.heuristics.heuristic import Heuristic
from nucs.memory import MAX, MIN


class VariableHeuristic(Heuristic):
    """
    Chooses a variable and one or several values for this variable.
    """

    def __init__(self, variable_heuristic: Callable, domain_heuristic: Callable):
        """
        Inits the heuristic.
        :param variable_heuristic: a function for choosing the variable
        :param domain_heuristic: a function for choosing a new domain for this variable
        """
        self.variable_heuristic = variable_heuristic
        self.domain_heuristic = domain_heuristic

    def choose(
        self, choice_points: List[NDArray], shr_domains: NDArray, shr_domain_changes: NDArray, dom_indices: NDArray
    ) -> None:
        var_idx = self.variable_heuristic(shr_domains, dom_indices)
        shr_domains_copy = shr_domains.copy(order="F")
        self.domain_heuristic(shr_domains, shr_domain_changes, shr_domains_copy, dom_indices[var_idx])
        choice_points.append(shr_domains_copy)


@njit("int16(int32[::1, :], uint16[:])", cache=True)
def first_not_instantiated_var_heuristic(shr_domains: NDArray, dom_indices: NDArray) -> int:
    """
    Chooses the first non instantiated variable.
    :param shr_domains: the shared domains of the problem
    :param dom_indices: the domain indices of the problem variables
    :return: the index of the variable
    """
    for var_idx, dom_index in enumerate(dom_indices):
        if shr_domains[dom_index, MIN] < shr_domains[dom_index, MAX]:
            return var_idx
    return -1  # cannot happen


@njit("int16(int32[::1, :], uint16[:])", cache=True)
def smallest_domain_var_heuristic(shr_domains: NDArray, dom_indices: NDArray) -> int:
    """
    Chooses the variable with the smallest domain and which is not instantiated.
    :param shr_domains: the shared domains of the problem
    :param dom_indices: the domain indices of the problem variables
    :return: the index of the variable
    """
    min_size = sys.maxsize
    min_idx = -1
    for var_idx, dom_index in enumerate(dom_indices):
        size = shr_domains[dom_index, MAX] - shr_domains[dom_index, MIN]  # actually this is size - 1
        if 0 < size < min_size:
            min_idx = var_idx
            min_size = size
    return min_idx


@njit("(int32[::1, :], boolean[::1, :], int32[::1, :], uint16)", cache=True)
def min_value_dom_heuristic(
    shr_domains: NDArray, shr_domain_changes: NDArray, shr_domains_copy: NDArray, domain_idx: int
) -> None:
    """
    Chooses the first value of the domain
    :param shr_domains: the shared domains of the problem
    :param: shr_domain_changes: the changes to the shared domains
    :param: shr_domains_copy: the shared domains to be added to the choice point
    :param domain_idx: the index of the domain
    """
    value = shr_domains[domain_idx, MIN]
    shr_domains_copy[domain_idx, MIN] = value + 1
    shr_domains[domain_idx, MAX] = value
    shr_domain_changes[domain_idx, MAX] = True


@njit("(int32[::1, :], boolean[::1, :], int32[::1, :], uint16)", cache=True)
def split_low_dom_heuristic(
    shr_domains: NDArray, shr_domain_changes: NDArray, shr_domains_copy: NDArray, domain_idx: int
) -> None:
    """
    Chooses the first half of the domain
    :param shr_domains: the shared domains of the problem
    :param: shr_domain_changes: the changes to the shared domains
    :param: shr_domains_copy: the shared domains to be added to the choice point
    :param domain_idx: the index of the domain
    """
    value = (shr_domains[domain_idx, MIN] + shr_domains[domain_idx, MAX]) // 2
    shr_domains_copy[domain_idx, MIN] = value + 1
    shr_domains[domain_idx, MAX] = value
    shr_domain_changes[domain_idx, MAX] = True



from nucs.problems.problem import Problem
from nucs.propagators.propagators import (
    ALG_AFFINE_LEQ,
    ALG_EXACTLY_EQ,
    ALG_LEXICOGRAPHIC_LEQ,
)


class SchurLemmaProblem(Problem):
    """
    See https://www.csplib.org/Problems/prob015/.
    """

    def __init__(self, n: int) -> None:
        super().__init__(shr_domains=[(0, 1)] * n * 3, dom_indices=list(range(n * 3)), dom_offsets=[0] * n * 3)
        propagators = []
        for x in range(n):
            propagators.append(([x * 3, x * 3 + 1, x * 3 + 2], ALG_EXACTLY_EQ, [1, 1]))
        for x in range(n):
            for y in range(n):
                z = (x + 1) + (y + 1) - 1
                if 0 <= z < n:
                    for k in range(3):
                        propagators.append(([3 * x + k, 3 * y + k, 3 * z + k], ALG_AFFINE_LEQ, [1, 1, 1, 2]))
        # breaking symmetries
        propagators.append(
            (list(range(0, n * 3, 3)) + list(range(1, n * 3, 3)) + list(range(2, n * 3, 3)), ALG_LEXICOGRAPHIC_LEQ, [])
        )
        self.set_propagators(propagators)



from typing import List, Tuple, Union

import numpy as np
from numpy.typing import NDArray

START = 0
END = 1

MIN = 0
MAX = 1

PROP_INCONSISTENCY = 0
PROP_CONSISTENCY = 1
PROP_ENTAILMENT = 2


def new_domains_by_values(domains: List[Union[int, Tuple[int, int]]]) -> NDArray:
    return np.array(
        [(domain, domain) if isinstance(domain, int) else domain for domain in domains],
        dtype=np.int32,
        order="F",
    )


def new_indices_by_values(dom_indices: List[int]) -> NDArray:
    return np.array(dom_indices, dtype=np.uint16, order="C")


def new_indices(n: int) -> NDArray:
    return np.empty(n, dtype=np.uint16, order="C")


def new_domain_offsets_by_values(dom_offsets: List[int]) -> NDArray:
    return np.array(dom_offsets, dtype=np.int32, order="C")


def new_offsets(n: int) -> NDArray:
    return np.empty(n, dtype=np.int32, order="C")


def new_data_by_values(data: List[int]) -> NDArray:
    return np.array(data, dtype=np.int32, order="C")


def new_data(n: int) -> NDArray:
    return np.empty(n, dtype=np.int32, order="C")


def new_domain_changes(n: int) -> NDArray:
    return np.ones((n, 2), dtype=bool, order="F")


def new_triggers(n: int, init_value: bool) -> NDArray:
    if init_value:
        return np.ones((n, 2), dtype=bool, order="F")
    else:
        return np.zeros((n, 2), dtype=bool, order="F")


def new_triggered_propagators(n: int) -> NDArray:
    return np.empty(n, dtype=bool, order="C")


def new_entailed_propagators(n: int) -> NDArray:
    return np.zeros(n, dtype=bool, order="C")


def new_algorithms(n: int) -> NDArray:
    return np.empty(n, dtype=np.int8, order="C")


def new_bounds(n: int) -> NDArray:
    return np.empty((n, 2), dtype=np.uint16)



from typing import List, Tuple

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_ALLDIFFERENT


class SudokuProblem(Problem):
    """
    A simple model for the sudoku problem.
    """

    def __init__(self, givens: List[List[int]]):
        super().__init__(
            shr_domains=[(1, 9) if given == 0 else (given, given) for line in givens for given in line],
            dom_indices=list(range(81)),
            dom_offsets=[0] * 81,
        )
        propagators: List[Tuple[List[int], int, List[int]]] = []
        for i in range(9):
            propagators.append((list(range(0 + i * 9, 9 + i * 9)), ALG_ALLDIFFERENT, []))
            propagators.append((list(range(0 + i, 81 + i, 9)), ALG_ALLDIFFERENT, []))
        for i in range(3):
            for j in range(3):
                offset = i * 27 + j * 3
                propagators.append(
                    (
                        [
                            0 + offset,
                            1 + offset,
                            2 + offset,
                            9 + offset,
                            10 + offset,
                            11 + offset,
                            18 + offset,
                            19 + offset,
                            20 + offset,
                        ],
                        ALG_ALLDIFFERENT,
                        [],
                    )
                )
        self.set_propagators(propagators)

    def pretty_print_solution(self, solution: List[int]) -> None:
        for i in range(0, 81, 9):
            print(solution[i : i + 9])



from nucs.problems.problem import Problem
from nucs.propagators.propagators import (
    ALG_EXACTLY_EQ,
    ALG_LEXICOGRAPHIC_LEQ,
    ALG_MIN_EQ,
)


class BIBDProblem(Problem):
    """
    A simple model for the BIBD problem.
    See https://www.csplib.org/Problems/prob028/ for a complete description of the problem.
    """

    def __init__(self, v: int, b: int, r: int, k: int, l: int):
        s = (v * (v - 1)) // 2
        n = (v + s) * b
        super().__init__(
            shr_domains=[(0, 1)] * n,
            dom_indices=list(range(n)),
            dom_offsets=[0] * n,
        )
        propagators = []
        # rows: counts
        for i in range(0, v):
            propagators.append((list(range(i * b, (i + 1) * b)), ALG_EXACTLY_EQ, [1, r]))
        # columns: counts
        for j in range(0, b):
            propagators.append((list(range(j, v * b, b)), ALG_EXACTLY_EQ, [1, k]))
        # scalar products: conjunctions and counts
        conj_idx = v * b
        for i1 in range(0, v - 1):
            for i2 in range(i1 + 1, v):
                conj_vars = []
                for j in range(0, b):
                    propagators.append(
                        ([(i1 * b + j), (i2 * b + j), conj_idx], ALG_MIN_EQ, [])
                    )  # TODO: replace by AND ?
                    conj_vars.append(conj_idx)
                    conj_idx += 1
                propagators.append((conj_vars, ALG_EXACTLY_EQ, [1, l]))
        # remove symmetries
        # lexleq on rows
        for i in range(0, v - 1):
            propagators.append((list(range(i * b, (i + 2) * b)), ALG_LEXICOGRAPHIC_LEQ, []))
        # lexleq on columns
        for j in range(0, b - 1):
            propagators.append((list(range(j, v * b, b)) + list(range(j + 1, v * b, b)), ALG_LEXICOGRAPHIC_LEQ, []))
        self.set_propagators(propagators)



from abc import abstractmethod
from typing import List

from numpy.typing import NDArray


class Heuristic:
    """
    Makes a choice.
    A choice can be more complex than just instantiating a variable.
    """

    @abstractmethod
    def choose(
        self, choice_points: List[NDArray], shr_domains: NDArray, shr_domain_changes: NDArray, dom_indices: NDArray
    ) -> None:
        """
        Makes a choice.
        :param choice_points: the choice point stack
        :param shr_domains: the shared domains of the problem
        :param shr_domain_changes: an array of shared domain changes
        :param dom_indices: the domain indices of the problem variables
        """
        pass



from typing import Iterator, List, Optional

from nucs.heuristics.heuristic import Heuristic
from nucs.heuristics.variable_heuristic import (
    VariableHeuristic,
    first_not_instantiated_var_heuristic,
    min_value_dom_heuristic,
)
from nucs.memory import new_domain_changes
from nucs.problems.problem import Problem, is_solved
from nucs.solvers.solver import Solver
from nucs.statistics import (
    STATS_OPTIMIZER_SOLUTION_NB,
    STATS_SOLVER_BACKTRACK_NB,
    STATS_SOLVER_CHOICE_DEPTH,
    STATS_SOLVER_CHOICE_NB,
    STATS_SOLVER_SOLUTION_NB,
)


class BacktrackSolver(Solver):
    """
    A solver relying on a backtracking mechanism.
    """

    def __init__(
        self,
        problem: Problem,
        heuristic: Heuristic = VariableHeuristic(first_not_instantiated_var_heuristic, min_value_dom_heuristic),
    ):
        super().__init__(problem)
        self.choice_points = []  # type: ignore
        self.shr_domain_changes = new_domain_changes(len(self.problem.shr_domains))
        self.heuristic = heuristic

    def solve(self) -> Iterator[List[int]]:
        while True:
            solution = self.solve_one()
            if solution is None:
                break
            yield solution
            if not self.backtrack():
                break

    def solve_one(self) -> Optional[List[int]]:
        """
        Find at most one solution.
        :return: the solution if it exists or None
        """
        while self.filter():
            if is_solved(self.problem.shr_domains):
                self.problem.statistics[STATS_SOLVER_SOLUTION_NB] += 1
                return self.problem.get_values()  # problem is solved
            self.heuristic.choose(
                self.choice_points,
                self.problem.shr_domains,
                self.shr_domain_changes,
                self.problem.dom_indices,
            )
            self.problem.statistics[STATS_SOLVER_CHOICE_NB] += 1
            self.problem.statistics[STATS_SOLVER_CHOICE_DEPTH] = max(
                self.problem.statistics[STATS_SOLVER_CHOICE_DEPTH], len(self.choice_points)
            )
        return None

    def minimize(self, variable_idx: int) -> Optional[List[int]]:
        solution = None
        while (new_solution := self.solve_one()) is not None:
            solution = new_solution
            self.problem.statistics[STATS_OPTIMIZER_SOLUTION_NB] += 1
            self.reset()
            self.problem.set_max_value(variable_idx, solution[variable_idx] - 1)
        return solution

    def backtrack(self) -> bool:
        """
        Backtracks and updates the problem's domains
        :return: true iff it is possible to backtrack
        """
        if len(self.choice_points) == 0:
            return False
        self.problem.statistics[STATS_SOLVER_BACKTRACK_NB] += 1
        self.shr_domain_changes[:, :] = True
        self.problem.shr_domains = self.choice_points.pop()
        self.problem.reset_entailed_propagators()
        return True

    def filter(self) -> bool:
        """
        Achieves bound consistency and backtracks if necessary.
        :return: true iff bound consistency has been achieved
        """
        while not self.problem.filter(self.shr_domain_changes):
            if not self.backtrack():
                return False
        return True

    def reset(self) -> None:
        """
        Resets the solver by resetting the problem and the choice points.
        """
        self.choice_points.clear()
        self.shr_domain_changes[:, :] = True
        self.problem.reset_shr_domains()
        self.problem.reset_entailed_propagators()



from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.memory import PROP_CONSISTENCY, new_triggers


def get_triggers(n: int, data: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :return: an array of triggers
    """
    return new_triggers(n, True)


@njit("int64(int32[::1,:], int32[:])", cache=True)
def compute_domains(domains: NDArray, data: NDArray) -> int:
    """
    A propagator that does nothing.
    """
    return PROP_CONSISTENCY



from typing import List, Tuple

import numpy as np
from numpy.typing import NDArray

from nucs.memory import MAX, MIN
from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AFFINE_EQ, ALG_AFFINE_LEQ, ALG_ALLDIFFERENT

GOLOMB_LENGTHS = [0, 0, 1, 3, 6, 11, 17, 25, 34, 44, 55, 72, 85, 106, 127]


def sum_first(n: int) -> int:
    return (n * (n + 1)) // 2


def index(mark_nb: int, i: int, j: int) -> int:
    return i * mark_nb - sum_first(i) + j - i - 1


def init_domains(var_nb: int, mark_nb: int) -> NDArray:
    domains = np.empty((var_nb, 2), dtype=np.int32, order="F")
    for i in range(0, mark_nb - 1):
        for j in range(i + 1, mark_nb):
            domains[index(mark_nb, i, j), MIN] = GOLOMB_LENGTHS[j - i + 1] if j - i + 1 < mark_nb else sum_first(j - i)
    domains[:, MAX] = 2**16
    return domains


class GolombProblem(Problem):
    """
    This is the famous Golomb ruler problem.
    It consists in finding n integers mark_i such that:
    - mark_0 = 0,
    - mark_0 <...< mark_n-1,
    - for all i<j, mark_j-mark_i are different,
    - mark_n-1 is minimal.
    Note: This is problem #6 in CSPLib.
    """

    def __init__(self, mark_nb: int) -> None:
        # dist_ij = mark_j - mark_i for j > i
        # mark_j = dist_0j for j > 0
        self.mark_nb = mark_nb
        self.length = mark_nb - 2
        var_nb = sum_first(mark_nb - 1)  # this the number of distances
        super().__init__(
            shr_domains=init_domains(var_nb, mark_nb),
            dom_indices=list(range(var_nb)),
            dom_offsets=[0] * var_nb,
        )
        propagators: List[Tuple[List[int], int, List[int]]] = []
        # redundant constraints
        for i in range(mark_nb - 1):
            for j in range(i + 1, mark_nb):
                if j - i < mark_nb - 1:
                    propagators.append(
                        (
                            [index(mark_nb, i, j), self.length],
                            ALG_AFFINE_LEQ,
                            [1, -1, -sum_first(mark_nb - 1 - (j - i))],
                        )
                    )
        # TODO break symmetries
        # main constraints
        for i in range(1, mark_nb - 1):
            for j in range(i + 1, mark_nb):
                propagators.append(
                    (
                        [index(mark_nb, 0, j), index(mark_nb, 0, i), index(mark_nb, i, j)],
                        ALG_AFFINE_EQ,
                        [1, -1, -1, 0],
                    )
                )
        propagators.append((list(range(var_nb)), ALG_ALLDIFFERENT, []))
        self.set_propagators(propagators)



from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_ALLDIFFERENT


class QueensProblem(Problem):
    """
    A simple model for the n-queens problem.
    """

    def __init__(self, n: int):
        super().__init__(
            shr_domains=[(0, n - 1)] * n,
            dom_indices=list(range(n)) * 3,
            dom_offsets=[0] * n + list(range(n)) + list(range(0, -n, -1)),
        )
        self.set_propagators(
            [
                (list(range(n)), ALG_ALLDIFFERENT, []),
                (list(range(n, 2 * n)), ALG_ALLDIFFERENT, []),
                (list(range(2 * n, 3 * n)), ALG_ALLDIFFERENT, []),
            ]
        )



from typing import Iterator, List, Optional

from nucs.problems.problem import Problem


class Solver:
    """
    A solver.
    """

    def __init__(self, problem: Problem):
        self.problem = problem

    def solve(self) -> Iterator[List[int]]:  # type: ignore
        """
        Returns an iterator over the solutions.
        :return: an iterator
        """
        pass

    def solve_all(self) -> List[List[int]]:
        """
        Returns the list of all solutions.
        :return: a list of list of integers
        """
        return [s for s in self.solve()]

    def find_all(self) -> None:
        """
        Find all solutions.
        """
        for _ in self.solve():
            pass

    def minimize(self, var_idx: int) -> Optional[List[int]]:  # type: ignore
        """
        Finds, if it exists, the solution to the problem that minimizes a given variable.
        :param variable_idx: the index of the variable
        :return: the solution if it exists or None
        """
        pass



import numpy as np
from numpy.typing import NDArray


def statistics_init() -> NDArray:
    """
    Inits a Numpy array for storing the statistics.
    :return: a Numpy array
    """
    return np.array([0] * STATS_MAX, dtype=np.int32, order="C")


def statistics_print(stats: NDArray) -> None:
    """
    Pretty-prints an array of statistics.
    :param stats: a Numpy array of statistics
    """
    print(
        {
            "OPTIMIZER_SOLUTION_NB": int(stats[STATS_OPTIMIZER_SOLUTION_NB]),
            "PROBLEM_FILTERS_NB": int(stats[STATS_PROBLEM_FILTER_NB]),
            "PROPAGATOR_ENTAILMENT_NB": int(stats[STATS_PROPAGATOR_ENTAILMENT_NB]),
            "PROPAGATOR_FILTERS_NB": int(stats[STATS_PROPAGATOR_FILTER_NB]),
            "PROPAGATOR_FILTERS_NO_CHANGE_NB": int(stats[STATS_PROPAGATOR_FILTER_NO_CHANGE_NB]),
            "PROPAGATOR_INCONSISTENCY_NB": int(stats[STATS_PROPAGATOR_INCONSISTENCY_NB]),
            "SOLVER_BACKTRACK_NB": int(stats[STATS_SOLVER_BACKTRACK_NB]),
            "SOLVER_CHOICE_NB": int(stats[STATS_SOLVER_CHOICE_NB]),
            "SOLVER_CHOICE_DEPTH": int(stats[STATS_SOLVER_CHOICE_DEPTH]),
            "SOLVER_SOLUTION_NB": int(stats[STATS_SOLVER_SOLUTION_NB]),
        }
    )


STATS_MAX = 10
STATS_OPTIMIZER_SOLUTION_NB = 0
STATS_PROBLEM_FILTER_NB = 1
STATS_PROPAGATOR_ENTAILMENT_NB = 2
STATS_PROPAGATOR_FILTER_NB = 3
STATS_PROPAGATOR_FILTER_NO_CHANGE_NB = 4
STATS_PROPAGATOR_INCONSISTENCY_NB = 5
STATS_SOLVER_BACKTRACK_NB = 6
STATS_SOLVER_CHOICE_NB = 7
STATS_SOLVER_CHOICE_DEPTH = 8
STATS_SOLVER_SOLUTION_NB = 9



from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.memory import END, MAX, MIN, START
from nucs.propagators import (
    affine_eq_propagator,
    affine_geq_propagator,
    affine_leq_propagator,
    alldifferent_propagator,
    count_eq_propagator,
    dummy_propagator,
    exactly_eq_propagator,
    lexicographic_leq_propagator,
    max_eq_propagator,
    max_leq_propagator,
    min_eq_propagator,
    min_geq_propagator,
)

# The ordinals of the algorithms for all propagators (sorted by alphabetical ordering).
ALG_AFFINE_EQ = 0
ALG_AFFINE_GEQ = 1
ALG_AFFINE_LEQ = 2
ALG_ALLDIFFERENT = 3
ALG_COUNT_EQ = 4
ALG_DUMMY = 5
ALG_EXACTLY_EQ = 6
ALG_LEXICOGRAPHIC_LEQ = 7
ALG_MAX_EQ = 8
ALG_MAX_LEQ = 9
ALG_MIN_EQ = 10
ALG_MIN_GEQ = 11

PROPAGATOR_MODULES = [
    affine_eq_propagator,
    affine_geq_propagator,
    affine_leq_propagator,
    alldifferent_propagator,
    count_eq_propagator,
    dummy_propagator,
    exactly_eq_propagator,
    lexicographic_leq_propagator,
    max_eq_propagator,
    max_leq_propagator,
    min_eq_propagator,
    min_geq_propagator,
]


def get_triggers(algorithm: int, size: int, data: NDArray) -> NDArray:
    """
    Returns the triggers for a propagator.
    :param algorithm: the ordinal of the algorithm
    :param size: the number of variables
    :return: an array of triggers
    """
    return PROPAGATOR_MODULES[algorithm].get_triggers(size, data)


@njit("int64(uint8, int32[::1,:], int32[:])", cache=True)
def compute_domains(algorithm: int, domains: NDArray, data: NDArray) -> int:
    """
    Computes the new domains for the variables.
    :param algorithm: the ordinal of the algorithm
    :param domains: the initial domains of the variables
    :param data: the parameters of the propagator
    :return: a status as an integer (INCONSISTENCY, CONSISTENCY, ENTAILMENT)
    """
    if algorithm == ALG_AFFINE_EQ:
        return affine_eq_propagator.compute_domains(domains, data)
    if algorithm == ALG_AFFINE_GEQ:
        return affine_geq_propagator.compute_domains(domains, data)
    if algorithm == ALG_AFFINE_LEQ:
        return affine_leq_propagator.compute_domains(domains, data)
    if algorithm == ALG_ALLDIFFERENT:
        return alldifferent_propagator.compute_domains(domains, data)
    if algorithm == ALG_COUNT_EQ:
        return count_eq_propagator.compute_domains(domains, data)
    if algorithm == ALG_DUMMY:
        return dummy_propagator.compute_domains(domains, data)
    if algorithm == ALG_EXACTLY_EQ:
        return exactly_eq_propagator.compute_domains(domains, data)
    if algorithm == ALG_LEXICOGRAPHIC_LEQ:
        return lexicographic_leq_propagator.compute_domains(domains, data)
    if algorithm == ALG_MAX_EQ:
        return max_eq_propagator.compute_domains(domains, data)
    if algorithm == ALG_MAX_LEQ:
        return max_leq_propagator.compute_domains(domains, data)
    if algorithm == ALG_MIN_EQ:
        return min_eq_propagator.compute_domains(domains, data)
    return min_geq_propagator.compute_domains(domains, data)


@njit(cache=True)
def init_triggered_propagators(
    triggered_propagators: NDArray,
    entailed_propagators: NDArray,
    shr_domain_changes: NDArray,
    propagator_nb: int,
    prop_var_bounds: NDArray,
    prop_dom_indices: NDArray,
    prop_triggers: NDArray,
) -> None:
    triggered_propagators.fill(False)
    for prop_idx in range(propagator_nb):
        if not entailed_propagators[prop_idx]:
            for var_idx in range(prop_var_bounds[prop_idx, START], prop_var_bounds[prop_idx, END]):
                dom_idx = prop_dom_indices[var_idx]
                if (shr_domain_changes[dom_idx, MIN] and prop_triggers[var_idx, MIN]) or (
                    shr_domain_changes[dom_idx, MAX] and prop_triggers[var_idx, MAX]
                ):
                    triggered_propagators[prop_idx] = True
                    break


@njit(cache=True)
def update_triggered_propagators(
    triggered_propagators: NDArray,
    entailed_propagators: NDArray,
    shr_domain_changes: NDArray,
    propagator_nb: int,
    prop_var_bounds: NDArray,
    prop_dom_indices: NDArray,
    prop_triggers: NDArray,
    previous_prop_idx: int,
) -> None:
    for prop_idx in range(propagator_nb):
        if not entailed_propagators[prop_idx] and prop_idx != previous_prop_idx and not triggered_propagators[prop_idx]:
            for var_idx in range(prop_var_bounds[prop_idx, START], prop_var_bounds[prop_idx, END]):
                dom_idx = prop_dom_indices[var_idx]
                if (shr_domain_changes[dom_idx, MIN] and prop_triggers[var_idx, MIN]) or (
                    shr_domain_changes[dom_idx, MAX] and prop_triggers[var_idx, MAX]
                ):
                    triggered_propagators[prop_idx] = True
                    break



import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.memory import (
    MAX,
    MIN,
    PROP_CONSISTENCY,
    PROP_ENTAILMENT,
    PROP_INCONSISTENCY,
    new_triggers,
)
from nucs.propagators.affine_eq_propagator import compute_domain_sum


def get_triggers(n: int, data: NDArray) -> NDArray:
    """
    Returns the triggers for this propagator.
    :param n: the number of variables
    :return: an array of triggers
    """
    triggers = new_triggers(n, False)
    for i in range(n):
        triggers[i, MIN] = data[i] > 0
        triggers[i, MAX] = data[i] < 0
    return triggers


@njit("int64(int32[::1,:], int32[:])", cache=True)
def compute_domains(domains: NDArray, a: NDArray) -> int:
    """
    Implements Sigma_i a_i * x_i <= a_{n-1}.
    :param domains: the domains of the variables
    :param a: the parameters of the propagator
    """
    n = len(domains)
    domain_sum = compute_domain_sum(n, domains, a)
    if domain_sum[MIN] >= 0:
        return PROP_ENTAILMENT
    new_domains = np.empty_like(domains)
    new_domains[:] = domains[:]
    for i in range(n):
        if a[i] > 0:
            new_max = domains[i, MIN] + (domain_sum[MAX] // a[i])
            new_domains[i, MAX] = min(domains[i, MAX], new_max)
        elif a[i] < 0:
            new_min = domains[i, MAX] - (-domain_sum[MAX] // a[i])
            new_domains[i, MIN] = max(domains[i, MIN], new_min)
        if new_domains[i, MIN] > new_domains[i, MAX]:
            return PROP_INCONSISTENCY
    domains[:] = new_domains[:]
    return PROP_CONSISTENCY



from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.memory import (
    MAX,
    MIN,
    PROP_CONSISTENCY,
    PROP_ENTAILMENT,
    PROP_INCONSISTENCY,
    new_triggers,
)


def get_triggers(n: int, data: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :return: an array of triggers
    """
    return new_triggers(n, True)


@njit("int64(int32[:,:], int32[:,:], int64, int64, int64, int64, int64)", cache=True)
def compute_domains_4(x: NDArray, y: NDArray, n: int, i: int, q: int, r: int, s: int) -> int:
    while i < n and x[i, MIN] == y[i, MAX]:
        i += 1
        # s = i
    if i < n and x[i, MIN] > y[i, MAX]:
        # xq < yq
        x[q, MAX] = min(x[q, MAX], y[q, MAX] - 1)
        if x[q, MAX] < x[q, MIN]:
            return PROP_INCONSISTENCY
        y[q, MIN] = max(y[q, MIN], x[q, MIN] + 1)
        if y[q, MAX] < y[q, MIN]:
            return PROP_INCONSISTENCY
        return PROP_ENTAILMENT if x[q, MAX] < y[q, MIN] else PROP_CONSISTENCY
    # u = 4
    return PROP_CONSISTENCY


@njit("int64(int32[:,:], int32[:,:], int64, int64, int64, int64, int64)", cache=True)
def compute_domains_3(x: NDArray, y: NDArray, n: int, i: int, q: int, r: int, s: int) -> int:
    while i < n and x[i, MAX] == y[i, MIN]:
        i += 1
        # s = i
    if i == n or x[i, MAX] < y[i, MIN]:
        # xq <= yq
        x[q, MAX] = min(x[q, MAX], y[q, MAX])
        if x[q, MAX] < x[q, MIN]:
            return PROP_INCONSISTENCY
        y[q, MIN] = max(y[q, MIN], x[q, MIN])
        if y[q, MAX] < y[q, MIN]:
            return PROP_INCONSISTENCY
        return PROP_ENTAILMENT if x[q, MAX] <= y[q, MIN] else PROP_CONSISTENCY
    # u = 3
    return PROP_CONSISTENCY


@njit("int64(int32[:,:], int32[:,:], int64, int64, int64, int64, int64)", cache=True)
def compute_domains_2(x: NDArray, y: NDArray, n: int, i: int, q: int, r: int, s: int) -> int:
    while i < n and x[i, MIN] == x[i, MAX] == y[i, MIN] == y[i, MAX]:
        i += 1
        r = i
    if i == n or x[i, MAX] < y[i, MIN]:
        # xq <= yq
        x[q, MAX] = min(x[q, MAX], y[q, MAX])
        if x[q, MAX] < x[q, MIN]:
            return PROP_INCONSISTENCY
        y[q, MIN] = max(y[q, MIN], x[q, MIN])
        if y[q, MAX] < y[q, MIN]:
            return PROP_INCONSISTENCY
        return PROP_ENTAILMENT if x[q, MAX] <= y[q, MIN] else PROP_CONSISTENCY
    if x[i, MIN] > y[i, MAX]:
        # xq < yq
        x[q, MAX] = min(x[q, MAX], y[q, MAX] - 1)
        if x[q, MAX] < x[q, MIN]:
            return PROP_INCONSISTENCY
        y[q, MIN] = max(y[q, MIN], x[q, MIN] + 1)
        if y[q, MAX] < y[q, MIN]:
            return PROP_INCONSISTENCY
        return PROP_ENTAILMENT if x[q, MAX] < y[q, MIN] else PROP_CONSISTENCY
    if x[i, MAX] == y[i, MIN] and x[i, MIN] < y[i, MAX]:
        if s > i + 1:
            i = s
        else:
            i += 1
            s = i
        return compute_domains_3(x, y, n, i, q, r, s)
    if x[i, MIN] == y[i, MAX] and x[i, MAX] > y[i, MIN]:
        if s > i + 1:
            i = s
        else:
            i += 1
            s = i
        return compute_domains_4(x, y, n, i, q, r, s)
    # u = 2
    return PROP_CONSISTENCY


@njit("int64(int32[:,:], int32[:,:], int64, int64, int64, int64, int64)", cache=True)
def compute_domains_1(x: NDArray, y: NDArray, n: int, i: int, q: int, r: int, s: int) -> int:
    while i < n and x[i, MIN] == y[i, MAX]:
        # enforce xi = yi
        x[i, MAX] = min(x[i, MAX], y[i, MAX])
        if x[i, MAX] < x[i, MIN]:
            return PROP_INCONSISTENCY
        y[i, MIN] = max(y[i, MIN], x[i, MIN])
        if y[i, MAX] < y[i, MIN]:
            return PROP_INCONSISTENCY
        i += 1
        q = i
    if i == n or x[i, MAX] < y[i, MIN]:
        return PROP_ENTAILMENT
    # enforce xq <= yq
    x[i, MAX] = min(x[i, MAX], y[i, MAX])
    if x[i, MAX] < x[i, MIN]:
        return PROP_INCONSISTENCY
    y[i, MIN] = max(y[i, MIN], x[i, MIN])
    if y[i, MAX] < y[i, MIN]:
        return PROP_INCONSISTENCY
    if r > i + 1:
        i = r
    else:
        i += 1
        r = i
    return compute_domains_2(x, y, n, i, q, r, s)


@njit("int64(int32[::1,:], int32[:])", cache=True)
def compute_domains(domains: NDArray, data: NDArray) -> int:
    """
    Implements lexicographic leq.
    See https://www.diva-portal.org/smash/record.jsf?pid=diva2:1041533.
    :param domains: the domains of the variables
    :param data: unused here
    """
    # TODO: make incremental
    n = len(domains) // 2
    return compute_domains_1(domains[:n], domains[n:], n, 0, 0, 0, 0)
